""" Module that contains test for the cyckic checking functionality of the
directed graph """

import unittest
from pythonalgos.graph.directed_graph import DirectedGraph
import os
import time
from pythonalgos.util.logging import Logging
from pythonalgos.util import path_tools as pt
from os import path


class TestDirectedGraphCyclic(unittest.TestCase):

    def setUp(self):
        pass

    def test_cyclic(self):
        Logging.enable()
        self.vertices = {0: [1], 1: [2, 3], 2: [3],
                         3: [4], 4: [5, 2], 5: [6], 6: [7], 7: [5]}
        self.directed_graph = DirectedGraph(self.vertices)
        self.assertTrue(self.directed_graph.is_cyclic())

    def test_acyclic(self):
        Logging.enable()
        self.vertices = {0: [1], 1: [2, 3], 2: [
            3], 3: [4, 6], 4: [5, 6], 5: [], 6: []}
        self.directed_graph = DirectedGraph(self.vertices)
        self.assertFalse(self.directed_graph.is_cyclic())

    def tearDown(self):
        pass


if __name__ == '__main__':
    unittest.main()



""" Module that contains test for the main directed graph functionality
"""

from pythonalgos.graph.directed_graph_core import DirectedGraphCore
import unittest
from pythonalgos.graph.directed_graph import DirectedGraph
from typing import Union


class TestDirectedGraph(unittest.TestCase):

    def setUp(self):
        self.vertices = {0: [1], 1: [2, 3], 2: [3],
                         3: [4, 6], 4: [5, 6], 5: [5], 6: [6]}
        self.directed_graph = DirectedGraph(self.vertices)

    def test_init(self):
        self.assertEqual(len(self.vertices.keys()),
                         self.directed_graph.get_vertices_count())

    def test_str(self):
        print(self.directed_graph)

    def test_one_vertex_self_loop(self):
        self.vertices = {0: [0]}
        self.directed_graph = DirectedGraph(self.vertices)

    def test_add_vertex(self):
        label = 7
        self.directed_graph.add_vertex(label)
        vertex = self.directed_graph.get_vertex(label)
        self.assertIsNotNone(vertex)
        self.assertEqual(vertex.get_outdegree(), 0)
        self.assertEqual(vertex.get_indegree(), 0)
        self.assertListEqual(vertex.get_heads(), list())

    def test_add_duplicate_vertex(self):
        label = 7
        self.directed_graph.add_vertex(label)
        with self.assertRaises(RuntimeError):
            self.directed_graph.add_vertex(label)

    def test_add_heads(self):
        vertex_to_test = 7
        self.directed_graph.add_vertex(vertex_to_test)
        vertex = self.directed_graph.get_vertex(vertex_to_test)
        no_heads = 3
        for i in range(no_heads):
            vertex.add_edge(self.directed_graph.get_vertex(i))
        self.assertEqual(len(vertex.get_heads()), no_heads)
        self.assertEqual(vertex.get_outdegree(), len(vertex.get_heads()))

    def test_outdegree(self):
        vertex = self.directed_graph.get_vertex(1)
        self.assertEqual(vertex.get_outdegree(), len(vertex.get_heads()))

    def test_indegree(self):
        vertex_to_test = 6
        self.vertices = {0: [1], 1: [2, 3], 2: [3], 3: [4, vertex_to_test],
                         4: [5, vertex_to_test], 5: [5], vertex_to_test: []}
        self.directed_graph = DirectedGraph(self.vertices)
        vertex = self.directed_graph.get_vertex(vertex_to_test)
        self.assertEqual(vertex.get_indegree(), 2)

    def test_get_reversed_graph(self):
        self.directed_graph = DirectedGraph(self.vertices)
        self.directed_graph.reversed(inplace=True)
        self.check_big_reversed_graph(self.directed_graph)

    def test_get_reversed_small_graph_inplace_false(self):
        self.vertices = {0: [1], 1: [2], 2: []}
        self.directed_graph = DirectedGraph(self.vertices)
        reversed_graph = self.directed_graph.reversed(inplace=False)
        self.assertTrue(reversed_graph !=
                        self.directed_graph.get_direct_graph_core())
        self.check_small_reversed_graph(reversed_graph)

    def test_get_reversed_small_graph_inplace_true(self):
        self.vertices = {0: [1], 1: [2], 2: []}
        self.directed_graph = DirectedGraph(self.vertices)
        reversed_graph = self.directed_graph.reversed(inplace=True)
        self.assertTrue(reversed_graph ==
                        self.directed_graph.get_direct_graph_core())
        self.check_small_reversed_graph(reversed_graph)

    def tearDown(self):
        pass

    def check_small_reversed_graph(self, reversed_graph):
        self.assertTrue(self.amount_of_tails2(reversed_graph, 2) == 1)
        self.assertTrue(self.has_vertex_label_as_head2(reversed_graph, 2, 1))
        self.assertTrue(self.amount_of_tails2(reversed_graph, 1) == 1)
        self.assertTrue(self.has_vertex_label_as_head2(reversed_graph, 1, 0))
        self.assertTrue(self.amount_of_tails2(reversed_graph, 0) == 0)

    def check_big_reversed_graph(self, reversed_graph):
        self.assertTrue(self.amount_of_tails2(reversed_graph, 6) == 3)
        self.assertTrue(self.has_vertex_label_as_head2(reversed_graph, 6, 4))
        self.assertTrue(self.has_vertex_label_as_head2(reversed_graph, 6, 6))
        self.assertTrue(self.has_vertex_label_as_head2(reversed_graph, 6, 3))
        self.assertTrue(self.amount_of_tails2(reversed_graph, 5) == 2)
        self.assertTrue(self.has_vertex_label_as_head2(reversed_graph, 5, 5))
        self.assertTrue(self.has_vertex_label_as_head2(reversed_graph, 5, 4))
        self.assertTrue(self.amount_of_tails2(reversed_graph, 4) == 1)
        self.assertTrue(self.has_vertex_label_as_head2(reversed_graph, 4, 3))
        self.assertTrue(self.amount_of_tails2(reversed_graph, 3) == 2)
        self.assertTrue(self.has_vertex_label_as_head2(reversed_graph, 3, 1))
        self.assertTrue(self.has_vertex_label_as_head2(reversed_graph, 3, 2))
        self.assertTrue(self.amount_of_tails2(reversed_graph, 2) == 1)
        self.assertTrue(self.has_vertex_label_as_head2(reversed_graph, 2, 1))
        self.assertTrue(self.amount_of_tails2(reversed_graph, 1) == 1)
        self.assertTrue(self.has_vertex_label_as_head2(reversed_graph, 1, 0))
        self.assertTrue(self.amount_of_tails2(reversed_graph, 0) == 0)

    def has_vertex_label_as_head(
            self,
            graph: DirectedGraphCore,
            vertex_label: Union[str, int],
            head_label: Union[str, int]) -> bool:
        return True if head_label in \
            {v.get_label() for v in
                graph.get_vertex(vertex_label).get_heads()} \
            else False

    def amount_of_heads(
            self,
            graph: DirectedGraphCore,
            vertex_label: Union[str, int]) -> int:
        return len(graph.get_vertex(vertex_label).get_heads())

    def has_vertex_label_as_head2(
            self,
            graph: DirectedGraph,
            vertex_label: Union[str, int],
            head_label: Union[str, int]) -> bool:
        return True if head_label in \
            {v.get_label() for v in
                graph.get_vertex(vertex_label).get_heads()} \
            else False

    def amount_of_tails2(
            self,
            graph: DirectedGraph,
            vertex_label: Union[str, int]) -> int:
        return len(graph.get_vertex(vertex_label).get_heads())


if __name__ == '__main__':
    unittest.main()



""" Module that defines a logging class that offers basic logging with indentation
and log yes/no switch
"""


class Logging:
    """ Class that is accessed in a static way. It contains functions for basic logging
    """

    logging = False
    inc_level = 0
    inc_size = 3

    @classmethod
    def enable(cls):
        Logging.logging = True

    @classmethod
    def disable(cls):
        Logging.logging = False

    @classmethod
    def log(cls, message, *args, inc=0):
        """ Function that logs a message

        Args:
            message: The message (with placeholders) to be logged
            args: The actual values that will replace the placeholders
            inc(bool). If True, first an indent will be perfromed. If false,
            no indent
        """

        if Logging.logging:
            if (inc != -1):
                Logging.inc_level += inc
            print(" " * (Logging.inc_level * Logging.inc_size) +
                  message.format(*args))
            if (inc == -1):
                Logging.inc_level -= 1

    @classmethod
    def inc_indent(cls):
        Logging.inc_level += 1

    @classmethod
    def dec_indent(cls):
        Logging.inc_level -= 1

    @classmethod
    def set_inc_size(cls, size):
        Logging.inc_size = size



""" Module that contains test for the main directed graph functionality
"""

from pythonalgos.graph.directed_graph_core import DirectedGraphCore
import unittest
from pythonalgos.graph.directed_graph import DirectedGraph
from typing import Union


class TestEdge(unittest.TestCase):

    def setUp(self):
        self.vertices = {0: [1], 1: [2], 2: []}
        self.directed_graph = DirectedGraph(self.vertices)

    def test_reverse_edge(self):
        vertex0 = self.directed_graph.get_vertex(0)
        vertex1 = self.directed_graph.get_vertex(1)
        vertex2 = self.directed_graph.get_vertex(2)
        edge = vertex1.get_edges()[0]
        edge.reverse()
        self.assertTrue(edge.get_tail() == vertex2)
        self.assertTrue(edge.get_head() == vertex1)

    def tearDown(self):
        pass


if __name__ == '__main__':
    unittest.main()



""" Module that contains the definition of a directed acyclic graph
"""

from pythonalgos.graph.directed_graph import DirectedGraph
from typing import Any, List, Mapping


class DirectedAcyclicGraph(DirectedGraph):
    """ Class to represent a directed acyclic graph. It inherits from
    DirectedGraph and when calling the __init__ constructor, it will check
    whether the graph adheres to the fact that the graph doesn't contain a
    cycle """

    def __init__(self, vertices: Mapping[Any, List[Any]]):
        """ Initializer that calls the super() initializer and then performs
        the cyclic test to check whether it is a directed acyclic graph """

        super().__init__(vertices)
        if super().is_cyclic():
            raise RuntimeError("Directed graph has a cycle")



""" Module that contains test for the strongly connected components
functionality of the directed graph
"""

import unittest
from pythonalgos.graph.directed_graph import DirectedGraph
from typing import List, Set
from pythonalgos.graph.vertex import Vertex


class TestDirectedGraphSCCs(unittest.TestCase):

    def setUp(self):
        pass

    def test_create_sccs_nontrivial(self):
        self.vertices = {0: [1], 1: [2, 3], 2: [3], 3: [4],
                         4: [5, 2], 5: [6], 6: [7], 7: [5], 8: [8]}
        self.directed_graph = DirectedGraph(self.vertices)
        sccs_labels = {frozenset(v.get_label() for v in s)
                       for s in self.directed_graph.create_sccs_kosaraju_dfs()}
        sccs_expected = {frozenset([2, 3, 4]), frozenset([5, 6, 7]),
                         frozenset([8])}
        self.assertTrue(sccs_expected == sccs_labels)

    def test_create_sccs_trivial(self):
        self.vertices = {0: [1], 1: [2, 3], 2: [0], 3: [4],
                         4: [5], 5: [3], 6: [5, 7], 7: [8],
                         8: [9], 9: [6, 10], 10: []}
        self.directed_graph = DirectedGraph(self.vertices)
        sccs_labels = {frozenset(v.get_label() for v in s)
                       for s in self.directed_graph.create_sccs_kosaraju_dfs(
                            nontrivial=False)}
        sccs_expected = \
            {frozenset([0, 1, 2]), frozenset([3, 4, 5]),
             frozenset([6, 7, 8, 9]), frozenset([10])}
        self.assertTrue(sccs_expected == sccs_labels)

    def tearDown(self):
        pass


if __name__ == '__main__':
    unittest.main()



class Advisor(object):

    def __init__(self):
        super().__init__()

    def advise(self, advice, *args, **kwargs):
        """ Method that retrieves the advice from the subclass and executes it

        Args:
            advice(str): The string that indicates the function in the subclass
        """

        try:
            advise_function = getattr(self, advice)
            advise_function(*args, **kwargs)
        except(AttributeError):
            # TODO Why doensn't the default work
            pass



import unittest
from pythonalgos.graph.directed_acyclic_graph import DirectedAcyclicGraph
import time
from pythonalgos.util.logging import Logging


class TestDirectedAcyclicGraph(unittest.TestCase):

    def test_init_ok(self):
        self.vertices = {0: [1], 1: [2, 3], 2: [3], 3: [4, 6], 4: [5, 6],
                         5: [], 6: []}
        DirectedAcyclicGraph(self.vertices)
        self.assertTrue(True)

    def test_init_not_ok(self):
        self.vertices = {0: [1], 1: [2, 3], 2: [3], 3: [4, 6, 1], 4: [5, 6],
                         5: [], 6: []}
        with self.assertRaises(RuntimeError):
            DirectedAcyclicGraph(self.vertices)

    def test_init_basic_not_ok(self):
        self.vertices = {0: [1], 1: [0]}
        with self.assertRaises(RuntimeError):
            DirectedAcyclicGraph(self.vertices)

    def test_init_basic_ok(self):
        self.vertices = {0: [1], 1: []}
        DirectedAcyclicGraph(self.vertices)

    def test_init_letters(self):
        self.vertices = {"0": ["1"], "1": ["2", "3"], "2": ["3"],
                         "3": ["4", "6"], "4": ["5", "6"],
                         "5": [], "6": []}
        DirectedAcyclicGraph(self.vertices)
        self.assertTrue(True)

    def test_init_booleans(self):
        self.vertices = {True: [False], False: []}
        DirectedAcyclicGraph(self.vertices)
        self.assertTrue(True)


if __name__ == '__main__':
    unittest.main()



from pythonalgos.graph.vertex import Vertex
from .. util.logging import Logging
from . directed_graph_core import DirectedGraphCore
from typing import Dict, Any, List, MutableMapping, Set

""" Module that contains the logic for kosaraju's SCCs algorithm
"""


def create_sccs_kosaraju_dfs(directed_graph: DirectedGraphCore,
                             nontrivial: bool) -> List[Set[Vertex]]:
    """ Function that creates a list of strongly connected components
    according to Kosaraju's algorithm
    (https://en.wikipedia.org/wiki/Kosaraju%27s_algorithm) with a
    depth-first-search approach.

    Args:
        directed_graph (DirectedGraph): The directed graph for which the SCCS
        should be calculated nontrivial(bool): If True, only nontrivial sccs
        will be returned, otherwise all sccs

    Returns:
        list(set()) of SCCs: Each SCC is a set of vertices

    """

    Logging.log("\nStarting")
    stack: List[Vertex] = list()
    sccs: List[Set[Vertex]] = list()
    visited: MutableMapping[Vertex, bool] = dict()
    for vertex in sorted(directed_graph.get_vertices(), key=lambda vertex: vertex.get_label()):
        if visited.get(vertex) is None:
            Logging.log("Vertex {0} not visited, go deep", vertex)
            fill_order_dfd_sccs(directed_graph, vertex, visited, stack)
        else:
            Logging.log("Vertex {0} already visited, skipping", vertex)

    visited = dict()
    directed_graph.reversed()
    for i in reversed(stack):
        if visited.get(i) is None:
            sccs.append(set())
            visit_dfs_sccs(directed_graph, i, visited,
                           sccs[-1])

    return filter_nontrivial(sccs, directed_graph) \
        if nontrivial else sccs


def filter_nontrivial(sccs_trivial: List[Set[Vertex]],
                      directed_graph: DirectedGraphCore) -> List[Set[Vertex]]:
    """ This function filters out the trivial sccs

    A scc is nontrivial, iff there are at least two vertices in it,
    or there is only one vertex with a self-loop. A self-loop means
    that the indegree and the outdegree are both 1 and the tail is equal
    to the head

    Args:
        sccs(list): The list of trivial sccs
        directed_graph(DirectedGraph): The directed graph

    Returns:
        sccs_nontrivial(list): The list of nontrivial sccs
    """

    sccs_non_trivial = list()
    for scc in sccs_trivial:
        vertex = next(iter(scc))
        if len(scc) >= 2 or \
            (len(scc) == 1 and vertex.get_indegree() == 1 and
                vertex.get_outdegree() == 1 and
                vertex.get_heads()[0] == vertex):
            sccs_non_trivial.append(scc)

    return sccs_non_trivial


def visit_dfs_sccs(directed_graph: DirectedGraphCore, vertex: Vertex,
                   visited: MutableMapping[Any, bool], scc: Set[Vertex]):
    """ Function that performs a recursive depth first search on the directed
    graph to check whether vertices have been visisted

    Args:
        directed_graph(DirectedGraph): The directed graph
        vertex (label): The current vertex
        visited (dict): A dictionary that maintains whether vertices have been
                        visited
        scc (set): The current scc being constructed

    """

    visited[vertex] = True
    scc.add(vertex)
    for head in sorted(vertex.get_heads(), key=lambda vertex: vertex.get_label()):
        if visited.get(head) is None:
            visit_dfs_sccs(directed_graph, head, visited, scc)


def fill_order_dfd_sccs(directed_graph: DirectedGraphCore, vertex: Vertex,
                        visited: MutableMapping[Vertex, bool],
                        stack: List[Vertex]):
    """ Function that covers the first part of the algorith by determining
    the order of vertices, traversing the graph with a depth first search,
    recursively.

    Args:
        directed_graph (DirectedGraph): The directed graph
        vertex(str): The current vertex
        visited (dict): A dictionary that maintains whether vertices have
        been visited
        stack (list): stack that will be processed, used to inverse the order

    """

    visited[vertex] = True
    for head in sorted(vertex.get_heads(), key=lambda vertex: vertex.get_label()):
        if visited.get(head) is None:
            fill_order_dfd_sccs(directed_graph, head, visited, stack)
    stack.append(vertex)
    # TODO Doesn't work, not needed?



from __future__ import annotations
from . edge import Edge
from typing import List
from typing import Mapping
from typing import Any
from . vertex_sort_order import VertexSortOrder

""" Module that contains the definition of a vertex in the context of a
directed graph """


class Vertex():
    """ Class to represent details about a vertex in the context of a directed
        graph, being the indegree, outdegree and the tails that are its
        successors. It inherits from the generic Vertex class """

    def __init__(self, label: str, **attrs):
        """ Initialises the vertex by adding some specifics

        Args:
            label(str): the label of the vertex
            **attrs: additional attributes that define the vertex
        """

        self._label = label
        self._edges: List[Edge] = list()
        self._indegree = 0
        self._attrs = attrs

    def add_edge(self, head_vertex: Vertex):
        """ This method adds an edge to the set of edges maintained by the
        vertex

        Args:
            head_vertex: the head vertex to be added

        """

        self._edges.append(Edge(self, head_vertex))

    def set_attr(self, attr: str, value: Any):
        self._attrs[attr] = value

    def get_attr(self, attr: str):
        return self._attrs.get(attr)

    def get_attrs(self) -> Mapping[str, Any]:
        return self._attrs

    def get_label(self) -> str:
        return self._label

    def increase_indegree(self):
        """ This method increases the indegree for the incumbent vertex """
        self._indegree += 1

    def decrease_indegree(self):
        """ This method decreases the indegree for the incumbent vertex """
        self._indegree -= 1

    def get_heads(self) -> List[Vertex]:
        return [e.get_head() for e in self._edges]

    def get_edges(self, vertex_sorting: VertexSortOrder =
                  VertexSortOrder.NATURAL) -> List[Edge]:
        if vertex_sorting == VertexSortOrder.NATURAL:
            return self._edges
        else:
            return sorted(self._edges,
                          key=lambda edge: edge.get_tail().get_label(),
                          reverse=vertex_sorting == VertexSortOrder.DESC)

    def remove_edges(self):
        self._edges = list()

    def get_indegree(self) -> int:
        return self._indegree

    def get_outdegree(self) -> int:
        return len(self._edges)

    def __str__(self):
        return str(self.get_label()) + ", outdegree: {}".format(self.get_outdegree()) + \
               ", indegree: {}".format(self.get_indegree()) + \
               ", heads: " + ",".join([str(tail.get_label())
                                      for tail in self.get_heads()])
